
// permutation, stored in std::array, based at one

template <int sz>
class permu_arr_1 {
    template <int tz> friend class permu_arr_1;

    std::array<int,sz> guts;

    void verify_guts () const {
        if (! permu_safe) return;
        for (int i {0}; i < sz; ++i) {
            int const gi {guts[i]};
            if (gi < base) throw fail
                {"permu_arr_1::verify ... element too small"};
            if (gi > sz) throw fail
                {"permu_arr_1::verify ... element too large"};
            for (int j {i + 1}; j < guts.size(); ++j) {
                int const gj {guts[j]};
                if (gj == gi) throw fail
                    {"permu_arr_1::verify ... element repeated"};
            }
        }
    }
public:
    static constexpr int base {1};

    permu_arr_1 (permu_arr_1 const &) = default;
    permu_arr_1 (permu_arr_1 &&) = default;
    permu_arr_1 & operator= (permu_arr_1 const &) = default;
    permu_arr_1 & operator= (permu_arr_1 &&) = default;
    ~permu_arr_1 () = default;

    explicit permu_arr_1 (bool const r = false) {
        std::iota (guts.begin(), guts.end(), base);
        if (! r) return;
        static std::random_device rd;
        static std::mt19937 temp (rd());
        std::shuffle (guts.begin(), guts.end(), temp);
        verify_guts ();
    }

    permu_arr_1 (std::initializer_list <int> i);

    template <typename input_iter>
        permu_arr_1 (input_iter begin, input_iter end);

    explicit permu_arr_1 (std::array<int,sz> const & g) :
        guts {g}
    {
        verify_guts ();
    }

    static permu_arr_1 partial (std::initializer_list <int> i);

    template <typename forward_iter>
        static permu_arr_1 partial (forward_iter begin, forward_iter end);

    #if USE_ARR_0
        explicit permu_arr_1 (permu_arr_0<sz> const & o);
    #endif
    #if USE_VEC_1
        explicit permu_arr_1 (permu_vec_1 const & o);
    #endif

    /////////////////

    permu_arr_1 operator* (permu_arr_1 const & o) const {
        std::array<int,sz> result;
        for (int i {0}; i < sz; ++i)
            result[i] = o.guts[guts[i] - 1];
        return permu_arr_1 {result};
    }

    permu_arr_1 invert () const {
        std::array<int,sz> result;
        for (int i {0}; i < sz; ++i) {
            int const gi {guts[i] - 1};
            for (int j {0}; j < sz; ++j) {
                int const gj {guts[j] - 1};
                if (gj == gi) result[gi] = j + 1;
            }
        }
        return permu_arr_1 {result};
    }

    template <int tz>
        permu_arr_1<sz+tz> direct_sum (permu_arr_1<tz> const & x) const;
    template <int tz>
        permu_arr_1<sz+tz> skew_sum (permu_arr_1<tz> const & x) const;

    template <int tz>
        permu_arr_1<tz> front_split () const;
    template <int tz>
        permu_arr_1<tz> back_split () const;

    template <int tz>
        std::pair <permu_arr_1<tz>, permu_arr_1<sz-tz>>
            front_split_both () const;
    template <int tz>
        std::pair <permu_arr_1<sz-tz>, permu_arr_1<tz>>
            back_split_both () const;

    bool front_splittable (int const tz) const;
    bool back_splittable (int const tz) const;

    template <int tz>
        permu_arr_1<sz*tz> laminate (permu_arr_1<tz> const & x) const;

    /////////////////

    bool next_lexi () {
        return std::next_permutation (guts.begin(), guts.end());
    }
    bool prev_lexi () {
        return std::prev_permutation (guts.begin(), guts.end());
    }
    bool next_SJT () {
        return SJT::next (guts, 1);
    }
    bool prev_SJT () {
        return SJT::prev (guts, 1);
    }

    void reverse_pos (int const lo = 1, int const hi = 0);
    void rotate_pos (int const lo, int const md, int const hi);
    void trans_pos (int const i, int const j, int const k = 1);

    void shuffle (bool const out_in, int const a, int const b) {
        if (permu_safe) {
            if (a < 1 ) throw fail {"permu_arr_1::shuffle ... a < 1 "};
            if (a > sz) throw fail {"permu_arr_1::shuffle ... a > sz"};
            if (b < 1 ) throw fail {"permu_arr_1::shuffle ... b < 1 "};
            if (b > sz) throw fail {"permu_arr_1::shuffle ... b > sz"};
        }
        shuffle::fore (guts, out_in, a - 1, b - 1);
        verify_guts ();
    }

    void unshuffle (bool const out_in, int const a, int const b) {
        if (permu_safe) {
            if (a < 1 ) throw fail {"permu_arr_1::unshuffle ... a < 1 "};
            if (a > sz) throw fail {"permu_arr_1::unshuffle ... a > sz"};
            if (b < 1 ) throw fail {"permu_arr_1::unshuffle ... b < 1 "};
            if (b > sz) throw fail {"permu_arr_1::unshuffle ... b > sz"};
        }
        shuffle::back (guts, out_in, a - 1, b - 1);
        verify_guts ();
    }

    /////////////////

    bool separable () const {
        return patterns::separable (guts);
    }

    bool stack_sortable () const {
        return patterns::stack_sortable (guts);
    }

    cycles_arr_1<sz> get_cycles () const {
        return cycles_arr_1<sz> {guts};
    }

    int inversion_count () const {
        int result {0};
        for (int i {0}; i < sz; ++i) {
            int const gi {guts[i]};
            for (int j {i + 1}; j < sz; ++j) {
                int const gj {guts[j]};
                if (gi > gj) ++result;
            }
        }
        return result;
    }

    int sign () const {
        return 1 - 2 * (inversion_count () % 2);
    }

    int order () const {
        permu_arr_1 const i (false);
        permu_arr_1 b {*this};

        int result {1};
        while (true) {
            if (b.guts == i.guts) break;
            b = (*this) * b;
            ++result;
        }
        return result;
    }

    bool is_deranged () const {
        for (int i {0}; i < sz; ++i)
            if (guts.at(i) == (i + 1)) return false;
        return true;
    }

    std::array<int,sz> const & at () const {
        return guts;
    }

    int at (int const i) const {
        return guts.at(i - base);
    }

    int seek (int const n) const {
        for (int i {0}; i < sz; ++i)
            if (guts[i - base] == n) return i;
        throw fail {"permu_arr_1::seek ... could not find"};
    }

    constexpr int size () const {
        return sz;
    }

    /////////////////

    void reverse_val (int const lo = 1, int const hi = 0);
    void rotate_val (int const lo, int const md, int const hi);
    void trans_val (int const i, int const j, int const k = 1);

    /////////////////

    void view (std::ostream & o) const {
        o << "[";
        for (auto const i: guts) o << " " << setw (2) << i;
        o << "  a1]";
    }
};

template <int sz>
std::ostream & operator<< (std::ostream & o, permu_arr_1<sz> const & p) {
    p.view (o);
    return o;
}

