
// permutation, stored in std::vector, based at one

class permu_vec_1 {
    std::vector<int> guts;

    void verify_guts () const {
        if (! permu_safe) return;
        int const sz {int (guts.size())};
        for (int i {0}; i < sz; ++i) {
            int const gi {guts[i]};
            if (gi < base) throw fail
                {"permu_vec_1::verify ... element too small"};
            if (gi > sz) throw fail
                {"permu_vec_1::verify ... element too large"};
            for (int j {i + 1}; j < guts.size(); ++j) {
                int const gj {guts[j]};
                if (gj == gi) throw fail
                    {"permu_vec_1::verify ... element repeated"};
            }
        }
    }

    void verify_size (permu_vec_1 const & o) const {
        if (! permu_safe) return;
        if (guts.size () != o.guts.size())
            throw fail {"permu_vec_1::verify_size ... mismatch"};
    }
public:
    static constexpr int base {1};

    permu_vec_1 () = delete;
    permu_vec_1 (permu_vec_1 const &) = default;
    permu_vec_1 (permu_vec_1 &&) = default;
    permu_vec_1 & operator= (permu_vec_1 const &) = default;
    permu_vec_1 & operator= (permu_vec_1 &&) = default;
    ~permu_vec_1 () = default;

    explicit permu_vec_1 (int const sz, bool const r = false) :
        guts (sz)
    {
        std::iota (guts.begin(), guts.end(), base);
        if (! r) return;
        static std::random_device rd;
        static std::mt19937 temp (rd());
        std::shuffle (guts.begin(), guts.end(), temp);
        verify_guts ();
    }

    permu_vec_1 (std::initializer_list <int> i) :
        guts {i}
    {
        verify_guts ();
    }

    template <typename input_iter>
    permu_vec_1 (input_iter begin, input_iter end) {
        for (input_iter i {begin}; i != end; ++i)
            guts.push_back (*i);
        verify_guts ();
    }

    explicit permu_vec_1 (std::vector<int> const & g) :
        guts {g}
    {
        verify_guts ();
    }

    static permu_vec_1 partial (int const sz, std::initializer_list <int> i);

    template <typename forward_iter>
        static permu_vec_1 partial (int const sz, forward_iter begin, forward_iter end);

    #if USE_VEC_0
        explicit permu_vec_1 (permu_vec_0 const & o);
    #endif
    #if USE_ARR_1
        template <int sz>
            explicit permu_vec_1 (permu_arr_1<sz> const & o);
    #endif

    /////////////////

    permu_vec_1 operator* (permu_vec_1 const & o) const {
        verify_size (o);
        int const sz {int (guts.size ())};
        std::vector<int> result (sz);
        for (int i {0}; i < sz; ++i)
            result[i] = o.guts[guts[i] - 1];
        return permu_vec_1 {result.begin(), result.end()};
    }    

    permu_vec_1 invert () const {
        int const sz {int (guts.size ())};
        std::vector<int> result (sz);
        for (int i {0}; i < sz; ++i) {
            int const gi {guts[i] - 1};
            for (int j {0}; j < sz; ++j) {
                int const gj {guts[j] - 1};
                if (gj == gi) result[gi] = j + 1;
            }
        }
        return permu_vec_1 {result.begin(), result.end()};
    }

    permu_vec_1 direct_sum (permu_vec_1 const & x) const;
    permu_vec_1 skew_sum (permu_vec_1 const & x) const;

    permu_vec_1 front_split (int const tz) const;
    permu_vec_1 back_split (int const tz) const;

    std::pair <permu_vec_1, permu_vec_1>
        front_split_both (int const tz) const;
    std::pair <permu_vec_1, permu_vec_1>
        back_split_both (int const tz) const;

    bool front_splittable (int const tz) const;
    bool back_splittable (int const tz) const;

    permu_vec_1 laminate (permu_vec_1 const & x) const;

    /////////////////

    bool next_lexi () {
        return std::next_permutation (guts.begin(), guts.end());
    }
    bool prev_lexi () {
        return std::prev_permutation (guts.begin(), guts.end());
    }
    bool next_SJT () {
        return SJT::next (guts, 1);
    }
    bool prev_SJT () {
        return SJT::prev (guts, 1);
    }

    void reverse_pos (int const lo = 1, int const hi = 0);
    void rotate_pos (int const lo, int const md, int const hi);
    void trans_pos (int const i, int const j, int const k = 1);

    void shuffle (bool const out_in, int const a, int const b) {
        int const sz {int (guts.size ())};
        if (permu_safe) {
            if (a < 1 ) throw fail {"permu_vec_1::shuffle ... a < 1 "};
            if (a > sz) throw fail {"permu_vec_1::shuffle ... a > sz"};
            if (b < 1 ) throw fail {"permu_vec_1::shuffle ... b < 1 "};
            if (b > sz) throw fail {"permu_vec_1::shuffle ... b > sz"};
        }
        shuffle::fore (guts, out_in, a - 1, b - 1);
        verify_guts ();
    }

    void unshuffle (bool const out_in, int const a, int const b) {
        int const sz {int (guts.size ())};
        if (permu_safe) {
            if (a < 1 ) throw fail {"permu_vec_1::unshuffle ... a < 1 "};
            if (a > sz) throw fail {"permu_vec_1::unshuffle ... a > sz"};
            if (b < 1 ) throw fail {"permu_vec_1::unshuffle ... b < 1 "};
            if (b > sz) throw fail {"permu_vec_1::unshuffle ... b > sz"};
        }
        shuffle::back (guts, out_in, a - 1, b - 1);
        verify_guts ();
    }

    /////////////////

    bool separable () const {
        return patterns::separable (guts);
    }

    bool stack_sortable () const {
        return patterns::stack_sortable (guts);
    }

    cycles_vec_1 get_cycles () const {
        return cycles_vec_1 {guts};
    }

    int inversion_count () const {
        int result {0};
        int const sz {int (guts.size ())};
        for (int i {0}; i < sz; ++i) {
            int const gi {guts[i]};
            for (int j {i + 1}; j < sz; ++j) {
                int const gj {guts[j]};
                if (gi > gj) ++result;
            }
        }
        return result;
    }

    int sign () const {
        return 1 - 2 * (inversion_count () % 2);
    }

    int order () const {
        int const sz {int (guts.size ())};
        permu_vec_1 const i (sz, false);
        permu_vec_1 b {*this};

        int result {1};
        while (true) {
            if (b.guts == i.guts) break;
            b = (*this) * b;
            ++result;
        }
        return result;
    }

    bool is_deranged () const {
        int const sz {int (guts.size ())};
        for (int i {0}; i < sz; ++i)
            if (guts.at(i) == (i + 1)) return false;
        return true;
    }

    std::vector<int> const & at () const {
        return guts;
    }

    int at (int const i) const {
        return guts.at(i - base);
    }

    int seek (int const n) const {
        int const sz {int (guts.size ())};
        for (int i {0}; i < sz; ++i)
            if (guts[i - base] == n) return i;
        throw fail {"permu_vec_1::seek ... could not find"};
    }

    int size () const {
        return int (guts.size());
    }

    /////////////////

    void reverse_val (int const lo = 1, int const hi = 0);
    void rotate_val (int const lo, int const md, int const hi);
    void trans_val (int const i, int const j, int const k = 1);

    /////////////////

    void view (std::ostream & o) const {
        o << "[";
        for (auto const i: guts) o << " " << setw (2) << i;
        o << "  v1]";
    }
};

std::ostream & operator<< (std::ostream & o, permu_vec_1 const & p) {
    p.view (o);
    return o;
}
