
namespace distance {
    template <typename permu>
    int cayley (permu const & a, permu const & b) {
        if (a.size() != b.size()) throw fail
            {"distance::cayley ... size mismatch"};
        auto const c (a * b.invert());
        auto const d (c.get_cycles().qty());
        return a.size() - d;
    }

    template <typename permu>
    int hamming (permu const & a, permu const & b) {
        if (a.size() != b.size()) throw fail
            {"distance::hamming ... size mismatch"};
        auto const & ag (a.at());
        auto const & bg (b.at());

        int result {0};
        for (int i {0}; i < a.size(); ++i) {
            if ((ag[i] - bg[i]) != 0)
                ++ result;
        }
        return result;
    }

    template <typename permu>
    double minkowski (permu const & a, permu const & b, double const p) {
        if (a.size() != b.size()) throw fail
            {"distance::minkowski ... size mismatch"};
        if (p <= 0.0) throw fail
            {"distance::minkowski ... p <= 0.0"};
        auto const & ag (a.at());
        auto const & bg (b.at());
        double sum {0.0};
        auto ai (ag.begin());
        auto bi (bg.begin());
        auto aj (ag.end());
        while (ai != aj)
            sum += std::pow (std::abs (*ai++ - *bi++), p);
        return std::pow (sum, 1.0 / p);
    }

    template <typename permu>
    int maximum (permu const & a, permu const & b) {
        if (a.size() != b.size()) throw fail
            {"distance::maximum ... size mismatch"};
        auto const & ag (a.at());
        auto const & bg (b.at());
        int curr {0};
        auto ai (ag.begin());
        auto bi (bg.begin());
        auto aj (ag.end());
        while (ai != aj) {
            int const cand {std::abs (*ai++ - *bi++)};
            if (cand > curr) curr = cand;
        }
        return curr;
    }
    
    template <typename permu>
    int minimum (permu const & a, permu const & b) {
        if (a.size() != b.size()) throw fail
            {"distance::minimum ... size mismatch"};
        auto const & ag (a.at());
        auto const & bg (b.at());
        int curr {a.size() + 1};
        auto ai (ag.begin());
        auto bi (bg.begin());
        auto aj (ag.end());
        while (ai != aj) {
            int const cand {std::abs (*ai++ - *bi++)};
            if (cand < curr) curr = cand;
        }
        return curr;
    }
}

namespace statistics {
    template <typename input_iter>
    double correlation_internal (
        input_iter const x_begin, input_iter const x_end,
        input_iter const y_begin, int const b
    ) {
        if (b < 0) throw fail {"statistics::correlation_internal ... b < 0"};
        if (b > 1) throw fail {"statistics::correlation_internal ... b > 1"};
        double n {0};
        double sum_xy {0.0};
        input_iter x {x_begin};
        input_iter y {y_begin};
        for (; x != x_end; ++n) sum_xy += (*x++) * (*y++);
        double const mean_xy {sum_xy / n};
        double const nb {n + b};
        double const sum_z {(nb * nb - nb) / 2.0};
        double const mean_z {sum_z / n};
        double const sum_zz  {(nb * (nb - 1.0) * (nb + nb - 1.0)) / 6.0};
        double const mean_zz {sum_zz / n};
        double const numer {mean_xy - mean_z * mean_z};
        double const denom {mean_zz - mean_z * mean_z};
        return numer / denom;
    }

    template <typename permu>
    double correlation (permu const & a, permu const & b) {
        if (a.size() != b.size()) throw fail
            {"statistics::correlation ... size mismatch"};
        return correlation_internal (
            a.at().begin(), a.at().end(),
            b.at().begin(), permu::base
        );
    }

    template <typename input_iter>
    double covariance_internal (
        input_iter const x_begin, input_iter const x_end,
        input_iter const y_begin, int const b
    ) {
        if (b < 0) throw fail {"statistics::covariance_internal ... b < 0"};
        if (b > 1) throw fail {"statistics::covariance_internal ... b > 1"};
        double n {0};
        double sum_xy {0.0};
        input_iter x {x_begin};
        input_iter y {y_begin};
        for (; x != x_end; ++n) sum_xy += (*x++) * (*y++);
        double const mean_xy {sum_xy / n};
        double const nb {n + b};
        double const sum_z {(nb * nb - nb) / 2.0};
        double const mean_z {sum_z / n};
        return mean_xy - mean_z * mean_z;
    }

    template <typename permu>
    double covariance (permu const & a, permu const & b) {
        if (a.size() != b.size()) throw fail
            {"statistics::covariance ... size mismatch"};
        return covariance_internal (
            a.at().begin(), a.at().end(),
            b.at().begin(), permu::base
        );
    }
};
